@@ -16,12 +16,18 @@ class SocketIOTestClient(object):
                       connects to the server on the global namespace.
     :param query_string: A string with custom query string arguments.
     :param headers: A dictionary with custom HTTP headers.
+    :param flask_test_client: The instance of the Flask test client
+                              currently in use. Passing the Flask test
+                              client is optional, but is necessary if you
+                              want the Flask user session and any other
+                              cookies set in HTTP routes accessible from
+                              Socket.IO events.
     """
     queue = {}
-    ack = None
+    acks = {}
 
     def __init__(self, app, socketio, namespace=None, query_string=None,
-                 headers=None):
+                 headers=None, flask_test_client=None):
         def _mock_send_packet(sid, pkt):
             if pkt.packet_type == packet.EVENT or \
                     pkt.packet_type == packet.BINARY_EVENT:
@@ -37,16 +43,23 @@ class SocketIOTestClient(object):
                                             'namespace': pkt.namespace or '/'})
             elif pkt.packet_type == packet.ACK or \
                     pkt.packet_type == packet.BINARY_ACK:
-                self.ack = {'args': pkt.data,
-                            'namespace': pkt.namespace or '/'}
+                self.acks[sid] = {'args': pkt.data,
+                                  'namespace': pkt.namespace or '/'}
+            elif pkt.packet_type in [packet.DISCONNECT, packet.ERROR]:
+                self.connected[pkt.namespace or '/'] = False
 
         self.app = app
+        self.flask_test_client = flask_test_client
         self.sid = uuid.uuid4().hex
         self.queue[self.sid] = []
+        self.acks[self.sid] = None
         self.callback_counter = 0
         self.socketio = socketio
+        self.connected = {}
         socketio.server._send_packet = _mock_send_packet
         socketio.server.environ[self.sid] = {}
+        socketio.server.async_handlers = False      # easier to test when
+        socketio.server.eio.async_handlers = False  # events are sync
         if isinstance(socketio.server.manager, PubSubManager):
             raise RuntimeError('Test client cannot be used with a message '
                                'queue. Disable the queue on your test '
@@ -55,6 +68,14 @@ class SocketIOTestClient(object):
         self.connect(namespace=namespace, query_string=query_string,
                      headers=headers)
 
+    def is_connected(self, namespace=None):
+        """Check if a namespace is connected.
+
+        :param namespace: The namespace to check. The global namespace is
+                         assumed if this argument is not provided.
+        """
+        return self.connected.get(namespace or '/', False)
+
     def connect(self, namespace=None, query_string=None, headers=None):
         """Connect the client.
 
@@ -76,12 +97,19 @@ class SocketIOTestClient(object):
             url += query_string
         environ = EnvironBuilder(url, headers=headers).get_environ()
         environ['flask.app'] = self.app
-        self.socketio.server._handle_eio_connect(self.sid, environ)
+        if self.flask_test_client:
+            # inject cookies from Flask
+            self.flask_test_client.cookie_jar.inject_wsgi(environ)
+        ret = self.socketio.server._handle_eio_connect(self.sid, environ)
+        if ret is None or ret is True:
+            self.connected['/'] = True
         if namespace is not None and namespace != '/':
             pkt = packet.Packet(packet.CONNECT, namespace=namespace)
             with self.app.app_context():
-                self.socketio.server._handle_eio_message(self.sid,
-                                                         pkt.encode())
+                ret = self.socketio.server._handle_eio_message(self.sid,
+                                                               pkt.encode())
+            if ret is None or ret is True:
+                self.connected[namespace] = True
 
     def disconnect(self, namespace=None):
         """Disconnect the client.
@@ -89,9 +117,12 @@ class SocketIOTestClient(object):
         :param namespace: The namespace to disconnect. The global namespace is
                          assumed if this argument is not provided.
         """
+        if not self.is_connected(namespace):
+            raise RuntimeError('not connected')
         pkt = packet.Packet(packet.DISCONNECT, namespace=namespace)
         with self.app.app_context():
             self.socketio.server._handle_eio_message(self.sid, pkt.encode())
+        del self.connected[namespace or '/']
 
     def emit(self, event, *args, **kwargs):
         """Emit an event to the server.
@@ -109,6 +140,8 @@ class SocketIOTestClient(object):
                           assumed if this argument is not provided.
         """
         namespace = kwargs.pop('namespace', None)
+        if not self.is_connected(namespace):
+            raise RuntimeError('not connected')
         callback = kwargs.pop('callback', False)
         id = None
         if callback:
@@ -116,7 +149,6 @@ class SocketIOTestClient(object):
             id = self.callback_counter
         pkt = packet.Packet(packet.EVENT, data=[event] + list(args),
                             namespace=namespace, id=id)
-        self.ack = None
         with self.app.app_context():
             encoded_pkt = pkt.encode()
             if isinstance(encoded_pkt, list):
@@ -124,9 +156,10 @@ class SocketIOTestClient(object):
                     self.socketio.server._handle_eio_message(self.sid, epkt)
             else:
                 self.socketio.server._handle_eio_message(self.sid, encoded_pkt)
-        if self.ack is not None:
-            return self.ack['args'][0] if len(self.ack['args']) == 1 \
-                else self.ack['args']
+        ack = self.acks.pop(self.sid, None)
+        if ack is not None:
+            return ack['args'][0] if len(ack['args']) == 1 \
+                else ack['args']
 
     def send(self, data, json=False, callback=False, namespace=None):
         """Send a text or JSON message to the server.
@@ -161,6 +194,8 @@ class SocketIOTestClient(object):
                           namespace is assumed if this argument is not
                           provided.
         """
+        if not self.is_connected(namespace):
+            raise RuntimeError('not connected')
         namespace = namespace or '/'
         r = [pkt for pkt in self.queue[self.sid]
              if pkt['namespace'] == namespace]
